package ru.landgrafhomyak.serdha.api.v0.ddl

import kotlin.jvm.JvmStatic
import kotlinx.datetime.Instant
import kotlinx.datetime.LocalDate
import kotlinx.datetime.LocalDateTime
import kotlinx.datetime.LocalTime

@Suppress("ClassName")
public abstract class ColumnType<RuntimeType : Any> {
	public abstract val typeName: String

	abstract override fun toString(): String

	public sealed class _StandardType<R : Any> : ColumnType<R>() {
		override fun toString(): String = "<serdha | column type '${this.typeName}'>"
	}

	public abstract class _VirtualType<R : Any>(public val wraps: ColumnType<R>) {
		override fun toString(): String = "<serdha | virtual column type 'ROW_ID'>"
	}

	public object ROW_ID : _StandardType<RowId<*>>() {
		override val typeName: String get() = "ROW_ID"
	}

	public object BOOLEAN : _StandardType<Boolean>() {
		override val typeName: String get() = "BOOLEAN"
	}

	public object INT_S8 : _StandardType<Byte>() {
		override val typeName: String get() = "INT_S8"
	}

	public object INT_U8 : _StandardType<UByte>() {
		override val typeName: String get() = "INT_U8"
	}

	public object INT_S16 : _StandardType<Short>() {
		override val typeName: String get() = "INT_S16"
	}

	public object INT_U16 : _StandardType<UShort>() {
		override val typeName: String get() = "INT_U16"
	}

	public object INT_S32 : _StandardType<Int>() {
		override val typeName: String get() = "INT_S32"
	}

	public object INT_U32 : _StandardType<UInt>() {
		override val typeName: String get() = "INT_U32"
	}

	public object INT_S64 : _StandardType<Long>() {
		override val typeName: String get() = "INT_S64"
	}

	public object INT_U64 : _StandardType<ULong>() {
		override val typeName: String get() = "INT_U64"
	}

	public object LOCAL_DATE : _StandardType<LocalDate>() {
		override val typeName: String get() = "LOCAL_DATE"
	}

	public object LOCAL_TIME : _StandardType<LocalTime>() {
		override val typeName: String get() = "LOCAL_TIME"
	}

	public object LOCAL_DATETIME : _StandardType<LocalDateTime>() {
		override val typeName: String get() = "LOCAL_DATETIME"
	}

	public object INSTANT : _StandardType<Instant>() {
		override val typeName: String get() = "INSTANT"
	}

	public object FLOAT_32 : _StandardType<Float>() {
		override val typeName: String get() = "FLOAT_32"
	}

	public object FLOAT_64 : _StandardType<Float>() {
		override val typeName: String get() = "FLOAT_64"
	}

	public class STRING internal constructor(
		private val _size: UInt
	) : _StandardType<String>() {
		override val typeName: String = if (this._size == 0u) "STRING" else "STRING(${this._size})"

		public val size: UInt? get() = if (this._size == 0u) null else this._size
	}

	public class BYTE_ARRAY internal constructor(
		private val _size: UInt
	) : _StandardType<String>() {
		override val typeName: String = if (this._size == 0u) "BYTE_ARRAY" else "BYTE_ARRAY(${this._size})"

		public val size: UInt? get() = if (this._size == 0u) null else this._size
	}

	public companion object DSL {
		@Suppress("RemoveRedundantQualifierName")
		@JvmStatic
		public val STRING: STRING = ColumnType.STRING(0u)

		@Suppress("RemoveRedundantQualifierName")
		@JvmStatic
		public fun STRING(size: UInt): STRING {
			if (size == 0u)
				throw IllegalArgumentException("Size of string can't be 0")
			return ColumnType.STRING(size)
		}

		@Suppress("RemoveRedundantQualifierName")
		@JvmStatic
		public val BYTE_ARRAY: BYTE_ARRAY = ColumnType.BYTE_ARRAY(0u)

		@Suppress("RemoveRedundantQualifierName")
		@JvmStatic
		public fun BYTE_ARRAY(size: UInt): BYTE_ARRAY {
			if (size == 0u)
				throw IllegalArgumentException("Size of byte array can't be 0")
			return ColumnType.BYTE_ARRAY(size)
		}

		@Suppress("FunctionName", "UNCHECKED_CAST", "RemoveRedundantQualifierName")
		@JvmStatic
		public fun <T : Any> ROW_ID(): ColumnType<RowId<T>> = ColumnType.ROW_ID as ColumnType<RowId<T>>
	}
}
